# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# ruff: noqa: A002

from collections.abc import Iterable
from typing import Any, overload

from pydiverse.common import (
    Bool,
    Date,
    Datetime,
    Dtype,
    Duration,
    Float,
    Int,
    List,
    String,
    Time,
)
from pydiverse.transform._internal import errors
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.tree import types
from pydiverse.transform._internal.tree.col_expr import (
    Col,
    ColExpr,
    ColFn,
    ColName,
    LiteralCol,
    WhenClause,
    wrap_literals,
)


def when(condition: ColExpr) -> WhenClause:
    condition = wrap_literals(condition)
    if condition.dtype() is not None and not types.without_const(condition.dtype()) == Bool():
        raise errors.DataTypeError(f"argument for `when` must be of boolean type, but has type `{condition.dtype()}`")

    return WhenClause([], wrap_literals(condition))


def lit(val: Any, dtype: Dtype | type[Dtype] | None = None) -> LiteralCol:
    """
    Creates a pydiverse.transform expression from a python builtin type.

    Usually, you can just use python builtins in expressions without wrapping them in
    ``lit``. The pydiverse.transform data type of the value is then inferred. However,
    ``lit`` allows to set the exact pydiverse.transform type, which may be useful
    sometimes.
    """
    errors.check_arg_type(Dtype | type | None, "lit", "dtype", dtype)
    if dtype is not None and not isinstance(dtype, Dtype):
        if not issubclass(dtype, Dtype) or dtype is List:
            raise TypeError(
                "argument for parameter `dtype` of `lit` must be an"
                "instance or subclass of `Dtype` (and for `List`, you have to"
                "specify the inner type)"
            )
        dtype = dtype()
    if dtype is not None and types.is_subtype(dtype):
        return LiteralCol(val, dtype).cast(dtype)
    return LiteralCol(val, dtype)


# --- from here the code is generated, do not delete this comment ---


def coalesce(arg: ColExpr, *args: ColExpr) -> ColExpr:
    """
    Returns the first non-null value among the given.

    :param arg:
        The first value.

    :param args:
        Further values. All must have the same type.

    Examples
    --------
    >>> t = pdt.Table(
    ...     {
    ...         "a": [5, None, 435, -1, 8, None],
    ...         "b": [-45, None, 6, 23, 1, 0],
    ...         "c": [10, 2, None, None, None, None],
    ...     }
    ... )
    >>> (
    ...     t
    ...     >> mutate(
    ...         x=pdt.coalesce(t.a, t.b, t.c),
    ...         y=pdt.coalesce(t.c, t.b, t.a),
    ...     )
    ...     >> show()
    ... )
    Table <unnamed>, backend: PolarsImpl
    shape: (6, 5)
    ┌──────┬──────┬──────┬─────┬─────┐
    │ a    ┆ b    ┆ c    ┆ x   ┆ y   │
    │ ---  ┆ ---  ┆ ---  ┆ --- ┆ --- │
    │ i64  ┆ i64  ┆ i64  ┆ i64 ┆ i64 │
    ╞══════╪══════╪══════╪═════╪═════╡
    │ 5    ┆ -45  ┆ 10   ┆ 5   ┆ 10  │
    │ null ┆ null ┆ 2    ┆ 2   ┆ 2   │
    │ 435  ┆ 6    ┆ null ┆ 435 ┆ 6   │
    │ -1   ┆ 23   ┆ null ┆ -1  ┆ 23  │
    │ 8    ┆ 1    ┆ null ┆ 8   ┆ 1   │
    │ null ┆ 0    ┆ null ┆ 0   ┆ 0   │
    └──────┴──────┴──────┴─────┴─────┘
    """

    return ColFn(ops.coalesce, arg, *args)


def count(
    *,
    partition_by: Col | ColName | str | Iterable[Col | ColName | str] | None = None,
    filter: ColExpr[Bool] | Iterable[ColExpr[Bool]] | None = None,
) -> ColExpr[Int]:
    """
    Returns the number of rows of the current table, like :code:`COUNT(*)` in SQL.
    """

    return ColFn(ops.count_star, partition_by=partition_by, filter=filter)


def dense_rank(
    *,
    partition_by: Col | ColName | str | Iterable[Col | ColName | str] | None = None,
    arrange: ColExpr | Iterable[ColExpr],
) -> ColExpr[Int]:
    """
    The number of smaller or equal values in the column (not counting duplicates).

    This function has two syntax alternatives, as shown in the example below. The
    pdt. version is a bit more flexible, because it allows sorting by multiple
    expressions.

    Examples
    --------
    >>> t = pdt.Table({"a": [5, -1, 435, -1, 8, None, 8]})
    >>> (
    ...     t
    ...     >> mutate(
    ...         x=t.a.nulls_first().dense_rank(),
    ...         y=pdt.dense_rank(arrange=t.a.nulls_first()),
    ...     )
    ...     >> show()
    ... )
    Table <unnamed>, backend: PolarsImpl
    shape: (7, 3)
    ┌──────┬─────┬─────┐
    │ a    ┆ x   ┆ y   │
    │ ---  ┆ --- ┆ --- │
    │ i64  ┆ i64 ┆ i64 │
    ╞══════╪═════╪═════╡
    │ 5    ┆ 3   ┆ 3   │
    │ -1   ┆ 2   ┆ 2   │
    │ 435  ┆ 5   ┆ 5   │
    │ -1   ┆ 2   ┆ 2   │
    │ 8    ┆ 4   ┆ 4   │
    │ null ┆ 1   ┆ 1   │
    │ 8    ┆ 4   ┆ 4   │
    └──────┴─────┴─────┘
    """

    return ColFn(ops.dense_rank, partition_by=partition_by, arrange=arrange)


def all(arg: ColExpr[Bool], *args: ColExpr[Bool]) -> ColExpr[Bool]:
    """ """

    return ColFn(ops.horizontal_all, arg, *args)


def any(arg: ColExpr[Bool], *args: ColExpr[Bool]) -> ColExpr[Bool]:
    """ """

    return ColFn(ops.horizontal_any, arg, *args)


@overload
def max(arg: ColExpr[Int], *args: ColExpr[Int]) -> ColExpr[Int]: ...


@overload
def max(arg: ColExpr[Float], *args: ColExpr[Float]) -> ColExpr[Float]: ...


@overload
def max(arg: ColExpr[String], *args: ColExpr[String]) -> ColExpr[String]: ...


@overload
def max(arg: ColExpr[Datetime], *args: ColExpr[Datetime]) -> ColExpr[Datetime]: ...


@overload
def max(arg: ColExpr[Time], *args: ColExpr[Time]) -> ColExpr[Time]: ...


@overload
def max(arg: ColExpr[Duration], *args: ColExpr[Duration]) -> ColExpr[Duration]: ...


@overload
def max(arg: ColExpr[Date], *args: ColExpr[Date]) -> ColExpr[Date]: ...


@overload
def max(arg: ColExpr[Bool], *args: ColExpr[Bool]) -> ColExpr[Bool]: ...


def max(arg: ColExpr, *args: ColExpr) -> ColExpr:
    """
    The maximum of the given columns.

    Examples
    --------
    >>> t = pdt.Table(
    ...     {
    ...         "a": [5, None, 435, -1, 8, None],
    ...         "b": [-45, None, 6, 23, -1, 0],
    ...         "c": [10, None, 2, None, -53, 3],
    ...     }
    ... )
    >>> t >> mutate(x=pdt.max(t.a, t.b, t.c)) >> show()
    Table <unnamed>, backend: PolarsImpl
    shape: (6, 4)
    ┌──────┬──────┬──────┬──────┐
    │ a    ┆ b    ┆ c    ┆ x    │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ i64  ┆ i64  │
    ╞══════╪══════╪══════╪══════╡
    │ 5    ┆ -45  ┆ 10   ┆ 10   │
    │ null ┆ null ┆ null ┆ null │
    │ 435  ┆ 6    ┆ 2    ┆ 435  │
    │ -1   ┆ 23   ┆ null ┆ 23   │
    │ 8    ┆ -1   ┆ -53  ┆ 8    │
    │ null ┆ 0    ┆ 3    ┆ 3    │
    └──────┴──────┴──────┴──────┘
    """

    return ColFn(ops.horizontal_max, arg, *args)


@overload
def min(arg: ColExpr[Int], *args: ColExpr[Int]) -> ColExpr[Int]: ...


@overload
def min(arg: ColExpr[Float], *args: ColExpr[Float]) -> ColExpr[Float]: ...


@overload
def min(arg: ColExpr[String], *args: ColExpr[String]) -> ColExpr[String]: ...


@overload
def min(arg: ColExpr[Datetime], *args: ColExpr[Datetime]) -> ColExpr[Datetime]: ...


@overload
def min(arg: ColExpr[Time], *args: ColExpr[Time]) -> ColExpr[Time]: ...


@overload
def min(arg: ColExpr[Duration], *args: ColExpr[Duration]) -> ColExpr[Duration]: ...


@overload
def min(arg: ColExpr[Date], *args: ColExpr[Date]) -> ColExpr[Date]: ...


@overload
def min(arg: ColExpr[Bool], *args: ColExpr[Bool]) -> ColExpr[Bool]: ...


def min(arg: ColExpr, *args: ColExpr) -> ColExpr:
    """
    The minimum of the given columns.

    Examples
    --------
    >>> t = pdt.Table(
    ...     {
    ...         "a": [5, None, 435, -1, 8, None],
    ...         "b": [-45, None, 6, 23, -1, 0],
    ...         "c": [10, None, 2, None, -53, 3],
    ...     }
    ... )
    >>> t >> mutate(x=pdt.min(t.a, t.b, t.c)) >> show()
    Table <unnamed>, backend: PolarsImpl
    shape: (6, 4)
    ┌──────┬──────┬──────┬──────┐
    │ a    ┆ b    ┆ c    ┆ x    │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ i64  ┆ i64  │
    ╞══════╪══════╪══════╪══════╡
    │ 5    ┆ -45  ┆ 10   ┆ -45  │
    │ null ┆ null ┆ null ┆ null │
    │ 435  ┆ 6    ┆ 2    ┆ 2    │
    │ -1   ┆ 23   ┆ null ┆ -1   │
    │ 8    ┆ -1   ┆ -53  ┆ -53  │
    │ null ┆ 0    ┆ 3    ┆ 0    │
    └──────┴──────┴──────┴──────┘
    """

    return ColFn(ops.horizontal_min, arg, *args)


@overload
def sum(arg: ColExpr[Int], *args: ColExpr[Int]) -> ColExpr[Int]: ...


@overload
def sum(arg: ColExpr[Float], *args: ColExpr[Float]) -> ColExpr[Float]: ...


@overload
def sum(arg: ColExpr[String], *args: ColExpr[String]) -> ColExpr[String]: ...


@overload
def sum(arg: ColExpr[Duration], *args: ColExpr[Duration]) -> ColExpr[Duration]: ...


def sum(arg: ColExpr, *args: ColExpr) -> ColExpr:
    """ """

    return ColFn(ops.horizontal_sum, arg, *args)


def rand() -> ColExpr[Float]:
    """
    Generates a column of random floating point number between 0 and 1.
    """

    return ColFn(ops.rand)


def rank(
    *,
    partition_by: Col | ColName | str | Iterable[Col | ColName | str] | None = None,
    arrange: ColExpr | Iterable[ColExpr],
) -> ColExpr[Int]:
    """
    The number of strictly smaller elements in the column plus one.

    This is the same as ``rank("min")`` in polars. This function has two syntax
    alternatives, as shown in the example below. The pdt. version is a bit more
    flexible, because it allows sorting by multiple expressions.


    Examples
    --------
    >>> t = pdt.Table({"a": [5, -1, 435, -1, 8, None, 8]})
    >>> (
    ...     t
    ...     >> mutate(
    ...         x=t.a.nulls_first().rank(),
    ...         y=pdt.rank(arrange=t.a.nulls_first()),
    ...     )
    ...     >> show()
    ... )
    Table <unnamed>, backend: PolarsImpl
    shape: (7, 3)
    ┌──────┬─────┬─────┐
    │ a    ┆ x   ┆ y   │
    │ ---  ┆ --- ┆ --- │
    │ i64  ┆ i64 ┆ i64 │
    ╞══════╪═════╪═════╡
    │ 5    ┆ 4   ┆ 4   │
    │ -1   ┆ 2   ┆ 2   │
    │ 435  ┆ 7   ┆ 7   │
    │ -1   ┆ 2   ┆ 2   │
    │ 8    ┆ 5   ┆ 5   │
    │ null ┆ 1   ┆ 1   │
    │ 8    ┆ 5   ┆ 5   │
    └──────┴─────┴─────┘
    """

    return ColFn(ops.rank, partition_by=partition_by, arrange=arrange)


def row_number(
    *,
    partition_by: Col | ColName | str | Iterable[Col | ColName | str] | None = None,
    arrange: ColExpr | Iterable[ColExpr] | None = None,
) -> ColExpr[Int]:
    """
    Computes the index of a row.

    Via the *arrange* argument, this can be done relative to a different order of
    the rows. But note that the result may not be unique if the argument of
    *arrange* contains duplicates.

    Examples
    --------
    >>> t = pdt.Table({"a": [5, -1, 435, -34, 8, None, 0]})
    >>> (
    ...     t
    ...     >> mutate(
    ...         x=pdt.row_number(),
    ...         y=pdt.row_number(arrange=t.a),
    ...     )
    ...     >> show()
    ... )
    Table <unnamed>, backend: PolarsImpl
    shape: (7, 3)
    ┌──────┬─────┬─────┐
    │ a    ┆ x   ┆ y   │
    │ ---  ┆ --- ┆ --- │
    │ i64  ┆ i64 ┆ i64 │
    ╞══════╪═════╪═════╡
    │ 5    ┆ 1   ┆ 5   │
    │ -1   ┆ 2   ┆ 3   │
    │ 435  ┆ 3   ┆ 7   │
    │ -34  ┆ 4   ┆ 2   │
    │ 8    ┆ 5   ┆ 6   │
    │ null ┆ 6   ┆ 1   │
    │ 0    ┆ 7   ┆ 4   │
    └──────┴─────┴─────┘
    """

    return ColFn(ops.row_number, partition_by=partition_by, arrange=arrange)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.transform._internal.util.warnings import warn

try:
    from sqlalchemy.ext.compiler import compiles
    from sqlalchemy.sql import Join

    # 1) DB2-specific compilation for FULL OUTER JOIN
    @compiles(Join, "ibm_db_sa")
    def _compile_join_db2(join, compiler, **kwargs):
        # If this is a FULL join, force FULL OUTER JOIN keyword
        kwargs = kwargs.copy()
        kwargs["asfrom"] = True
        if getattr(join, "full", False):
            return "".join(
                (
                    compiler.process(join.left, **kwargs),
                    " FULL OUTER JOIN ",
                    compiler.process(join.right, **kwargs),
                    " ON ",
                    compiler.process(join.onclause, **kwargs),
                )
            )
        # Otherwise, fall back to default behavior for LEFT/INNER
        return compiler.visit_join(join, **kwargs)
except ImportError:
    warn("Failed to patch SQLAlchemy FULL OUTER JOIN for DB2. ")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.transform as pdt
from pydiverse.transform._internal.pipe.c import C
from pydiverse.transform._internal.pipe.verbs import mutate, select
from pydiverse.transform.common import *
from tests.fixtures.backend import skip_backends
from tests.util.assertion import assert_result_equal


def test_exp(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.clip(-700, 700).exp() for c in t}),
    )


def test_log(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: pdt.max(1e-16, c).log() for c in t}))


def test_abs(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.abs() for c in t}),
    )


def test_round(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.round() for c in t}),
    )
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.round(2) for c in t}),
    )
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.round(-2) for c in t}),
    )


def test_add(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{f"add_{c.name}_{d.name}": c + d for d in t for c in t}),
    )


def test_sub(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{f"sub_{c.name}_{d.name}": c - d for d in t for c in t}),
    )


def test_neg(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{f"neg_{c.name}": -c for c in t}),
    )


def test_mul(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{f"mul_{c.name}_{d.name}": c * d for d in t for c in t}),
    )


def test_div(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(**{c.name: pdt.when(c.abs() < 1e-50).then(1e-50).otherwise(c) for c in t})
        >> (lambda s: mutate(**{f"div_{c.name}_{d.name}": c / d for d in s for c in s})),
    )


def test_decimal(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(f=t.f.cast(pdt.Decimal), g=t.g.cast(pdt.Decimal)) >> mutate(u=C.f + C.g, z=C.f * C.g),
    )


def test_floor(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.floor() for c in t}),
    )


def test_ceil(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.ceil() for c in t}),
    )


# MSSQL is really bad with inf/nan handling. IBM DB2 supports it with DECFLOAT but
# sqlalchemy doesn't support that.
@skip_backends("mssql", "ibm_db2")
def test_inf_lit(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> select()
        >> mutate(
            inf=float("inf"),
            neg_inf=float("-inf"),
        )
        >> mutate(
            inf_str=C.inf.cast(pdt.String()),
            neg_inf_str=C.neg_inf.cast(pdt.String()),
        )
        >> mutate(
            inf_back=C.inf_str.cast(pdt.Float64()),
            neg_inf_back=C.neg_inf_str.cast(pdt.Float64()),
        ),
    )


# MSSQL is really bad with inf/nan handling. IBM DB2 supports it with DECFLOAT but
# sqlalchemy doesn't support that.
@skip_backends("mssql", "sqlite", "ibm_db2")
def test_nan_lit(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> select()
        >> mutate(nan=float("nan"))
        >> mutate(nan_str=C.nan.cast(pdt.String()))
        >> mutate(nan_back=C.nan_str.cast(pdt.Float64())),
    )


# MSSQL is really bad with inf/nan handling. IBM DB2 supports it with DECFLOAT but
# sqlalchemy doesn't support that.
@skip_backends("mssql", "ibm_db2")
def test_is_inf(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(inf=float("inf"))
        >> mutate(neg_inf=-C.inf)
        >> mutate(
            inf_is_inf=C.inf.is_inf(),
            neg_inf_is_inf=C.neg_inf.is_inf(),
            **{c.name + "is_inf": c.is_inf() for c in t},
            inf_is_not_inf=C.inf.is_not_inf(),
            neg_inf_is_not_inf=C.neg_inf.is_not_inf(),
            **{c.name + "is_not_inf": c.is_not_inf() for c in t},
        ),
    )


# MSSQL is really bad with inf/nan handling. IBM DB2 supports it with DECFLOAT but
# sqlalchemy doesn't support that.
@skip_backends("mssql", "sqlite", "ibm_db2")
def test_is_nan(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(nan=float("nan"))
        >> mutate(
            nan_is_nan=C.nan.is_nan(),
            **{c.name + "is_nan": c.is_nan() for c in t},
            nan_is_not_nan=C.nan.is_not_nan(),
            **{c.name + "is_not_nan": c.is_not_nan() for c in t},
        ),
    )


def test_int_pow(df_int):
    assert_result_equal(df_int, lambda t: t >> mutate(u=pdt.min(t.a, 10) ** pdt.min(t.b.abs(), 5)))


def test_sin(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.sin() for c in t}))


def test_cos(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.cos() for c in t}))


def test_tan(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.atan().tan() for c in t}))


def test_asin(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.sin().asin() for c in t}))


def test_acos(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.cos().acos() for c in t}))


def test_atan(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.atan() for c in t}))


def test_sqrt(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.abs().sqrt() for c in t}))


def test_cbrt(df_num):
    assert_result_equal(df_num, lambda t: t >> mutate(**{c.name: c.cbrt() for c in t}))


def test_log10(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: pdt.max(1e-16, c).log10() for c in t}),
    )


def test_clip(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(
            **{c.name + "1": c.clip(0.1, 10) for c in t},
            **{c.name + "2": c.clip(-500, 1024) for c in t},
        ),
    )


def test_rand(df_num):
    # just see that it runs
    def f(t):
        t >> mutate(r=pdt.rand()) >> export(Polars)
        return pdt.Table({"a": [1]})

    assert_result_equal(df_num, f)
